/*
 * Copyright © 2011 Linaro Limited
 *
 * This file is part of glcompbench.
 *
 * glcompbench is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * glcompbench is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with glcompbench.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  Alexandros Frantzis <alexandros.frantzis@linaro.org>
 *  Jesse Barker <jesse.barker@linaro.org>
 */

#include <string>
#include <sstream>
#include <cstdlib>

#include "composite-canvas-pixman.h"
#include "composite-window-pixman.h"
#include "options.h"
#include "log.h"

#include <sys/shm.h>

CompositeCanvasPixman::~CompositeCanvasPixman()
{
    release_resources();
}

bool
CompositeCanvasPixman::have_xshm()
{
    int ignore, major, minor;
    Bool pixmaps;
    bool have_xshm = false;
        
    /* Check for the XShm extension */
    if (XQueryExtension(xdpy_, "MIT-SHM", &ignore, &ignore, &ignore)) {
        if (XShmQueryVersion(xdpy_, &major, &minor, &pixmaps) == True) {
            Log::debug("XShm extension version %d.%d %s shared pixmaps\n",
                       major, minor, (pixmaps == True) ? "with" : "without");
            have_xshm = true;
        }
    }

    if (!have_xshm && Options::use_accel_tfp) {
        Log::info("*** Accelerated TFP requested but X Server doesn't support XSHM\n",
                  "    Using normal XImage transfer mechanism.");
    }

    return have_xshm;
}

bool
CompositeCanvasPixman::ensure_ximage_shm()
{
    XWindowAttributes canvas_attr;

    if (!XGetWindowAttributes(xdpy_, canvas_, &canvas_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    ximage_ = XShmCreateImage(xdpy_, canvas_attr.visual, canvas_attr.depth,
                              ZPixmap, 0, &shm_info_, width_, height_);
    if (!ximage_) {
        Log::error("XCreateImage failed!\n");
        return false;
    }
    
    /* Get shared memory from the system */
    shm_info_.shmid = shmget(IPC_PRIVATE, ximage_->bytes_per_line * ximage_->height,
                             IPC_CREAT | 0777);
    if (shm_info_.shmid < 0) {
        Log::error("shmget failed!\n");
        release_resources();
        return false;
    }

    /* Attach shared memory to process address space and set it as RW */
    shm_info_.readOnly = False;
    shm_info_.shmaddr = (char *)shmat(shm_info_.shmid, 0, 0);
    if (shm_info_.shmaddr == reinterpret_cast<char *>(-1)) {
        Log::error("shmat failed!\n");
        release_resources();
        return false;
    }

    /* Set shared memory as the pixel data for the XImage*/
    ximage_->data = shm_info_.shmaddr;
        
    /* Attach shared memory to X Server */
    if (!XShmAttach(xdpy_, &shm_info_)) {
        Log::error("XShmAttach failed!\n");
        release_resources();
        return false;
    }

    return true;
}

bool
CompositeCanvasPixman::ensure_ximage()
{
    XWindowAttributes canvas_attr;

    if (!XGetWindowAttributes(xdpy_, canvas_, &canvas_attr)) {
        Log::error("XGetWindowAttributes failed!\n");
        return false;
    }

    char *pixel_data = reinterpret_cast<char *>(calloc(4 * width_ * height_, 1));

    ximage_ = XCreateImage(xdpy_, canvas_attr.visual, canvas_attr.depth,
                           ZPixmap, 0, pixel_data, width_, height_, 32, 0);
    if (!ximage_) {
        Log::error("XCreateImage failed!\n");
        return false;
    }

    return true;
}

bool
CompositeCanvasPixman::ensure_pixman_image()
{
    if (pixman_image_)
        return true;

    bool have_ximage;

    if (use_shm_)
        have_ximage = ensure_ximage_shm();
    else
        have_ximage = ensure_ximage();

    if (!have_ximage)
        return false;

    gc_ =  XCreateGC(xdpy_, canvas_, 0, 0);
    if (!gc_) {
        Log::error("XCreateGC failed!\n");
        release_resources();
        return false;
    }

    pixman_image_ = pixman_image_create_bits(PIXMAN_a8r8g8b8, width_, height_,
                                             reinterpret_cast<uint32_t *>(ximage_->data),
                                             4 * width_);
    if (!pixman_image_) {
        Log::error("XCreateImage failed!\n");
        release_resources();
        return false;
    }

    return true;
}

XVisualInfo *
CompositeCanvasPixman::get_canvas_xvisualinfo()
{
    int num_visuals;
    XVisualInfo vis_tmpl;
    XVisualInfo *vis_info;

    vis_tmpl.depth = 32;
    vis_tmpl.bits_per_rgb = 8;
    vis_info = XGetVisualInfo(xdpy_, VisualDepthMask | VisualBitsPerRGBMask,
                              &vis_tmpl, &num_visuals);
    if (!vis_info) {
        Log::error("Error: couldn't get X visual\n");
        return 0;
    }

    return vis_info;
}

void
CompositeCanvasPixman::reshape(int width, int height)
{
    CompositeCanvas::reshape(width, height);

    release_resources();
    ensure_pixman_image();
}

bool
CompositeCanvasPixman::make_current()
{
    use_shm_ = have_xshm() && Options::use_accel_tfp;

    return ensure_pixman_image();
}

void
CompositeCanvasPixman::release_resources()
{
    if (pixman_image_)
        pixman_image_unref(pixman_image_);

    if (gc_)
        XFreeGC(xdpy_, gc_);

    if (ximage_)
        XDestroyImage(ximage_);

    if (shm_info_.shmaddr != reinterpret_cast<char *>(-1)) {
        shmdt(shm_info_.shmaddr);
        XShmDetach(xdpy_, &shm_info_);
    }

    pixman_image_ = 0;
    gc_ = 0;
    ximage_ = 0;
    shm_info_.shmaddr = reinterpret_cast<char *>(-1);
}

CompositeWindow *
CompositeCanvasPixman::create_window(Window win)
{
    return new CompositeWindowPixman(xdpy_, win);
}

void
CompositeCanvasPixman::update_screen()
{
    if (use_shm_) {
        XShmPutImage(xdpy_, canvas_, gc_, ximage_, 0, 0, 0, 0, width_,
                     height_, False);
    }
    else {
        XPutImage(xdpy_, canvas_, gc_, ximage_, 0, 0, 0, 0, width_, height_);
    }
}

std::string
CompositeCanvasPixman::info_string()
{
    std::stringstream ss;

    ss << "    Pixman " << pixman_version_string() << std::endl;

    return ss.str();
}
